package sandbox;

import java.util.Iterator;

import processing.core.*;
import sandbox.Utils;

public class Camera implements PConstants {

    protected PApplet parent;
    protected Sandbox sandbox;
   
    public PVector camOrigin, camHead;
    protected float cameraDistance;

    protected boolean controlCollide;
    protected boolean controlColl;
    public PVector controlPos, controlHead;
    protected PVector controlPosFut, controlHeadFut;
    protected float controlRadius;
    protected int controlFcColl;
    protected float controlInterpolationFactor;
    protected PVector inertia;
   
    protected float vel, ang;
    protected float vMIN = (float) 1.;
    protected float vMAX = (float) 5.;
    protected float vACCEL = (float) .15;
    protected float vDACCEL = (float) .5;   

   
    public Camera(Sandbox sandbox_) {
        sandbox = sandbox_;
        parent = sandbox.parent;
       
        controlCollide = true;
        controlPos = new PVector(0, -1, 0);
        controlPosFut = new PVector(0, 0, 0);
       
        controlHead = new PVector(1, 0, 0);   
        controlHeadFut = new PVector(1, 0, 0);
        controlHead.set(Utils.toSpherical(controlHead));  
        controlHeadFut.set(Utils.toSpherical(controlHeadFut));
        controlRadius = 70;           
        controlInterpolationFactor = .03f;
        inertia = new PVector(0,0,0);
        vel = 0;
        ang = 0;       
       
        camOrigin = new PVector();
        camHead = new PVector();
       
        if (sandbox.mode == sandbox.FIRSTPERSON) {
            camHead.set(1,0,1);
            camOrigin.set(0, -50, 0);
        }
        else {     
            if (sandbox.isOPENGL) {
                camOrigin.set(0,-50,0);
                camHead.set(1000.f*(float).5, 700.f*(float).35, -1000.f*(float).35);
            }
            else {
                camOrigin.set(-1400,-1100,1000);
                camHead.set(-camOrigin.x, -camOrigin.y/11.f*7.f, -camOrigin.z);
            }
        }
       
        cameraDistance = 100f;
    }
   
   
   
    protected void head(float x_, float y_, float z_) {
        camHead.set(x_, y_, z_);
    }
   
   
   
    protected void anchor() {   
        parent.camera(camOrigin.x, camOrigin.y, camOrigin.z,
                camHead.x, camHead.y, camHead.z,
                0, 1, 0);        
        sandbox.cameraX = camOrigin.x;
        sandbox.cameraY = camOrigin.y;
        sandbox.cameraZ = camOrigin.z;
    }

   
   
    protected void place(Node node) {
        if (sandbox.controlMode == sandbox.MOUSE_SELECT) {
            controlPosFut.set(node.position.x, node.position.y, node.position.z);
           
            controlHeadFut.set(Utils.toCartesian(controlHeadFut));
            controlHeadFut.set(node.heading.x, node.heading.y, node.heading.z);
            controlHeadFut.set(Utils.toSpherical(controlHeadFut));
           
        }
    }

   
    protected void setup() {
        if (sandbox.mode == sandbox.THIRDPERSON) {
            if (sandbox.isOPENGL) {
                camOrigin.set(0,-50,0);
                camHead.set(1000.f*(float).5, 700.f*(float).35, -1000.f*(float).35);
            }
            else {
                camOrigin.set(-1400,-1100,1000);
                camHead.set(-camOrigin.x, -camOrigin.y/11.f*7.f, -camOrigin.z);
            }
        }
        else {
            camOrigin.set(0,-50,0);
        }
       
       
        if (sandbox.cameraMode == sandbox.TRACKING) {
            if (sandbox.mode == sandbox.FIRSTPERSON) {
                camOrigin.set(Utils.toSpherical(camOrigin));
                camOrigin.x = cameraDistance;
                camOrigin.y += parent.PI;
                camOrigin.z = -parent.QUARTER_PI;
                camOrigin.set(Utils.toCartesian(camOrigin));
            }
        }
        else if (sandbox.cameraMode == sandbox.FACING) {
            if (sandbox.mode == sandbox.FIRSTPERSON) {
                camOrigin.set(Utils.toSpherical(camOrigin));
                camOrigin.x = cameraDistance;
                camOrigin.z = -.1f;
                camOrigin.set(Utils.toCartesian(camOrigin));
            }
        }       
    }
   
   
    protected void isometricSpace() {   
        PVector controlPos0 = new PVector(controlPos.x, controlPos.y, controlPos.z);  
        PVector controlPosCollTest =
            Utils.interpolateTo(controlPos, controlPosFut, (float)0.06);
        controlPosCollTest.y = controlPosFut.y - 1;
       
        if (sandbox.controlMode == sandbox.MOUSE_PRESSED && controlCollide
            && sandbox.testCollisions((Node)null, controlPosCollTest)) {
            if (!sandbox.lastCollidedNode.walkable
                    && sandbox.lastCollidedNode.intersectsSphere(controlPosCollTest, controlRadius*.9f)) {
                inertia.add(PVector.mult(
                  sandbox.lastCollidedNode.intersectionNormal(controlPos), vMAX*2.f));
            }
            controlPos.add(inertia);
            inertia.mult(.9f);
           
            controlPosFut.set(Utils.interpolateTo(controlPosFut, controlPos, (float)0.003));
            controlPosFut.y = controlPos.y;
        }
        else {
            controlPosFut.add(inertia);
            inertia.mult(.9f);
           
            controlPos.set(Utils.interpolateTo(controlPos, controlPosFut, (float)0.03));
            controlPos.y = controlPosFut.y;
           
            if (sandbox.controlMode == sandbox.MOUSE_PRESSED &&
                sandbox.cameraMode == sandbox.TRACKING) {
                PVector velocity = PVector.sub(controlPos, controlPos0);
                PVector velocityFut = PVector.sub(controlPosFut, controlPos);
               
                if (!Utils.isZeroVector(velocity)) {
                    ang = Utils.headingXZ(velocity);   
                }
                if (parent.abs(ang-controlHead.y)>parent.PI) {
                    if (ang > controlHead.y) ang-=parent.TWO_PI;
                    else controlHead.y-=parent.TWO_PI;
                }   
                controlHead.y += .03*(ang-controlHead.y);    
            }  
            else if (sandbox.controlMode == sandbox.MOUSE_SELECT &&
                     (sandbox.cameraMode == sandbox.TRACKING ||
                         sandbox.cameraMode == sandbox.FACING)) {
                ang = controlHeadFut.y;   
                if (parent.abs(ang-controlHead.y) > parent.PI) {
                    if (ang > controlHead.y) ang -= parent.TWO_PI;
                    else controlHead.y -= parent.TWO_PI;
                }   
                controlHead.y += .03*(ang-controlHead.y);    
            }
        }   
    }
   

   
    protected void perspectiveSpace() {
        if (sandbox.controlMode == sandbox.MOUSE_PRESSED) {
            if (sandbox.myMousePressed)
                vel = parent.constrain(vel + vACCEL, vMIN, vMAX);
            else
                vel = parent.constrain(vel - vDACCEL, 0, vMAX);
           
        //     COLLISIONS   

            PVector normal = new PVector(0,0,0);

            if (sandbox.onFloor) controlPos.y = -1;  //TODO : feo feo
            int nColl = -1;
           
            Iterator i = sandbox.Nodes.values().iterator();
            while (i.hasNext()) {
                Node tmpNode = (Node)i.next();

                if (!controlColl && controlCollide && tmpNode.collides) {
                    if (tmpNode.intersectsSphere(controlPos, controlRadius)) {
                        if (tmpNode.walkable && sandbox.onFloor) {
                            controlPos.y = tmpNode.position.y-tmpNode.heightOver()-1;//tmpNode.bounds.getExtent().y-1;
                        }
                        else if (tmpNode.intersectsSphere(controlPos, .9f*controlRadius)) {
                            inertia.add(PVector.mult(
                                    tmpNode.intersectionNormal(controlPos),vMAX*2.f));
                        }
                        else {
                            normal.add(tmpNode.intersectionNormal(controlPos));
                        }
                    } 
                }
            }
            if (nColl == -1) controlColl = false;   
           
           
           
            // TODO: feo feo
            int sign = 1;
            if (sandbox.cameraMode == sandbox.TRACKING) sign = -1;
           
            PVector velocity = new PVector(vel, 0, 0);
            velocity.set(Utils.rotateY(velocity, -sign*controlHead.y));
                       
//            PVector velocity = new PVector(vel*scLUT.cos(-(ang+(float)2.5*parent.atan2(-(parent.mouseX-parent.width*(float).5), parent.width*(float).5))),
//                    0,
//                    vel*scLUT.sin(-(ang+(float)2.5*parent.atan2(-(parent.mouseX-parent.width*(float).5), parent.width*(float).5))));                              
           
//            PVector velocity = new PVector(vel*Utils.cos(-sign*controlHead.y),
//                                           0,
//                                           vel*Utils.sin(-sign*controlHead.y));
           
            if (vel != 0) {
                if (!Utils.isZeroVector(normal)) {
                    float normComp = velocity.dot(normal);
                    velocity.sub(PVector.mult(normal, normComp));
                }
               
                ang += sign*(float).1*parent.atan2(-(sandbox.myMouseX-parent.width*(float).5),
                                                parent.width*(float).5);
                controlPos.add(velocity); 
            }
            controlPos.add(inertia);
            inertia.mult(0.9f);
           
            controlHead.y += .3*(ang-controlHead.y);
           
           
        }
        else if (sandbox.controlMode == sandbox.MOUSE_SELECT) {
            controlPos.set(
                    Utils.interpolateTo(controlPos, controlPosFut, controlInterpolationFactor));
            float h = .01f*controlPos.dist(controlPosFut);
            controlPos.y -= h;
           
            float tmpAng = controlHeadFut.y;
            if (parent.abs(tmpAng-controlHead.y) > parent.PI) {
                if (tmpAng > controlHead.y) tmpAng -= parent.TWO_PI;
                else controlHead.y -= parent.TWO_PI;
            }   
            controlHead.y += .03*(tmpAng-controlHead.y);                
        }
    }
}
