package sandbox;

import processing.core.*;
import java.util.UUID;

public abstract class Node {
	  
	  Sandbox sandbox;
	  protected PVector position;
	  protected PVector heading;
	  protected PMatrix3D rotMatrix;
	  protected PMatrix3D invRotMatrix;
	  protected int[] nodeId;
	  protected String nodeHash;
	  protected UUID uuid;

	  protected float ang;
	  protected boolean inUse;
	  protected boolean mouseOver;
	  protected boolean collides;
	  protected boolean walkable;
	  protected boolean collision;
	  
	  protected boolean isBox;

	  
	  protected Node(Sandbox sandbox_, String nodeHash_) {
	    sandbox = sandbox_;
	    nodeHash = nodeHash_;
	    uuid = UUID.randomUUID();
	    position = new PVector(0, 0, 0);
	    heading = new PVector(0, 0, 1);
	    rotMatrix = new PMatrix3D();
	    invRotMatrix = new PMatrix3D();
	    nodeId = new int[7];
	    for (int n = 0; n < 7; n++) nodeId[n] = 0;

	    sandbox.Nodes.put(nodeHash, (Node)this);
	    mouseOver = false;
	    walkable = false;
	    collides = false;
	    collision = false;
	    inUse = true;
	    
	    isBox = false;
	  } 
	  
	  
	  protected void setPosition (float x_, float y_, float z_) {
	      position.x = x_;
	      position.y = y_;
	      position.z = z_;    
	  } 
	  
	  
	  protected void setMatrix(PMatrix3D mtx) {
		  rotMatrix.set(mtx);
		  setPosition(rotMatrix.m03, rotMatrix.m13, rotMatrix.m23);
		  invRotMatrix.set(rotMatrix);
		  invRotMatrix.invert();
		  heading.set(invRotMatrix.m20, invRotMatrix.m21, invRotMatrix.m22);		  
	  }
	  
	  
	  protected void setHash(String hash_) {
		  sandbox.Nodes.remove(nodeHash);
		  nodeHash = hash_;
		  sandbox.Nodes.put(nodeHash, (Node)this);
	  }
	  
	  
	  protected void setNodeId(int type_, float x_, float y_, float z_, 
			  								float lx_, float ly_, float lz_) {
		  nodeId[0] = type_;
		  nodeId[1] = sandbox.gridX(x_);
		  nodeId[2] = sandbox.gridY(y_);
		  nodeId[3] = sandbox.gridZ(z_);
		  nodeId[4] = sandbox.gridX(lx_);
		  nodeId[5] = sandbox.gridY(ly_);
		  nodeId[6] = sandbox.gridZ(lz_);		  
	  }
	  
	  
	  
	  protected boolean intersectsNode(Node node_) {
		  if (this.isBox && node_.isBox) {
			  NodeBox n1 = (NodeBox)this;
			  NodeBox n2 = (NodeBox)node_;
			  boolean test1 = n1.testSeparationFromOBB(n2);
			  boolean test2 = n2.testSeparationFromOBB(n1);
			  
			  return !(test1 || test2);
		  }
		  else return false;
	  }
	  
	  
	  protected abstract boolean intersectsSphere(PVector v_, float r_);
	  
	  
	  protected abstract PVector intersectionNormal(PVector v_);
	   
	  
	  protected abstract boolean containsVector(float x_, float y_, float z_);
	  
	  
	  protected abstract boolean containsVector(PVector v_);

	  
	  protected abstract float heightOver();
	  
	  
	  protected abstract void drawBounds();
	  
	  protected void drawCollisionArea() {
		  int fillColor = sandbox.parent.g.fillColor;
		  int strokeColor = sandbox.parent.g.strokeColor;
		  float strokeWeight = sandbox.parent.g.strokeWeight;		

		  sandbox.parent.stroke(50);    
		  sandbox.parent.strokeWeight(1);
		  sandbox.parent.noFill();		  
		  sandbox.parent.pushMatrix();
		  sandbox.parent.translate(position.x, position.y, position.z);
		  sandbox.parent.box(2*sandbox.RESOLUTION*sandbox.COLLISION_AREA);      
		  sandbox.parent.popMatrix();

		  sandbox.parent.stroke(strokeColor);        
		  sandbox.parent.strokeWeight(strokeWeight);		
		  sandbox.parent.fill(fillColor);		
	  } 

}