package sandbox;

import processing.core.*;


public class NodeBox extends Node {

	protected AABB bounds;

	NodeBox(Sandbox sandbox_, String nodeHash_) {
		super(sandbox_, nodeHash_);
		isBox = true;
		bounds = new AABB(new PVector(0, 0, 0), new PVector(0, 0, 0));
	}

	protected void setPosition(float x_, float y_, float z_) {
		super.setPosition(x_, y_, z_);
		bounds.set(x_, y_, z_);
	}

	protected void setExtent(float x_, float y_, float z_) {  
		bounds.setExtent(new PVector(x_, y_, z_));
	}

	protected boolean intersectsBox(NodeBox node_) {
		return bounds.intersectsBox(node_.bounds);
	}

	protected boolean intersectsSphere(PVector v_, float r_) {
		PVector innPos = inOBBSpace(v_.x, v_.y, v_.z);
		boolean out = bounds.intersectsSphere(innPos, r_);
		return out;
	}

	protected PVector intersectionNormal(PVector v_) {
		PVector normal = new PVector();
		PVector pnormal = new PVector();

		PVector innPos = inOBBSpace(v_.x, v_.y, v_.z);
		innPos.sub(position);

		float x = bounds.getExtent().x;
		float y = bounds.getExtent().y;
		float z = bounds.getExtent().z;

		float dx0 = (x-Math.abs(innPos.x)); 
		float dy0 = (y-Math.abs(innPos.y));
		float dz0 = (z-Math.abs(innPos.z));

		float minD = Math.min(dx0, Math.min(dy0, dz0));

		if (minD == dx0) {
			if (x*innPos.x < 0) pnormal.set(-1,0,0);
			else pnormal.set(1,0,0);
		}
		else if (minD == dy0) {
			if (y*innPos.y < 0) pnormal.set(0,-1,0);
			else pnormal.set(0,1,0);
		}
		else {
			if (z*innPos.z < 0) pnormal.set(0,0,-1);
			else pnormal.set(0,0,1);
		}

		PVector rotPnormal = new PVector();
		rotPnormal = rotMatrix.mult(pnormal, rotPnormal);
		normal.set(rotPnormal.x, rotPnormal.y, rotPnormal.z);
		normal.sub(position);
		return normal;
	}
	
	
	
	protected boolean testSeparationFromOBB(NodeBox nodeBox) {
		// nodeBox as OBB; this as AABB
		float[] x, y, z;
		float[] xRotW, yRotW, zRotW;
		float[] xRotO, yRotO, zRotO;
		x = new float[8]; y = new float[8]; z = new float[8];
		xRotW = new float[8]; yRotW = new float[8]; zRotW = new float[8];
		xRotO = new float[8]; yRotO = new float[8]; zRotO = new float[8];
		
		PVector distance = new PVector(nodeBox.position.x,
                					 	nodeBox.position.y,
                					 	nodeBox.position.z);  
		distance.sub(this.position);
		
		for (int n = 0; n < 8; n++) {
			x[n] = ((n%2 == 0)?1:-1) * nodeBox.bounds.extent.x;
			y[n] = ((n%4 > 1)?1:-1)  * nodeBox.bounds.extent.y;
			z[n] = ((n > 3)?1:-1)    * nodeBox.bounds.extent.z;
		}
		
		for (int n = 0; n < 8; n++) {
			xRotW[n] = nodeBox.rotMatrix.multX(x[n], y[n], z[n]);
			yRotW[n] = nodeBox.rotMatrix.multY(x[n], y[n], z[n]);
			zRotW[n] = nodeBox.rotMatrix.multZ(x[n], y[n], z[n]);
		}	
		
		for (int n = 0; n < 8; n++) {
			xRotO[n] = this.invRotMatrix.multX(xRotW[n], yRotW[n], zRotW[n]);
			yRotO[n] = this.invRotMatrix.multY(xRotW[n], yRotW[n], zRotW[n]);
			zRotO[n] = this.invRotMatrix.multZ(xRotW[n], yRotW[n], zRotW[n]);
		}


		for (int i = 0; i < 3; i++) {
			for (int j = 0; j < 2; j++) {
				boolean allPositive = true;
				for (int n = 0; n < 8; n++) {
					float a = 0, b = 0;
					switch(i) {
					case 0: a = this.bounds.extent.x;
							b = xRotO[n]; 
							break;
					case 1: a = this.bounds.extent.y;
							b = yRotO[n];
							break;
					case 2: a = this.bounds.extent.z;
							b = zRotO[n];
							break;
					}
					if (j == 0) {
						if (a >= b) {
							allPositive = false;
						}
					}
					else if (-a <= b) {
							allPositive = false;
					}
				}
				if (allPositive) return true;
			}
		}
		return false;
	}
	
	

	protected boolean containsVector(float x_, float y_, float z_) {
		PVector innPos = inOBBSpace(x_, y_, z_);	
		AABB expandedAABB = bounds.copy();
		expandedAABB.setExtent(PVector.mult(bounds.getExtent(), (float)1.2));		
		boolean out = expandedAABB.containsVector(innPos);
		return out;
	}	  

	protected boolean containsVector(PVector v_) {	
		return containsVector(v_.x, v_.y, v_.z);
	}	

	// TODO: Optimization: VectorPool for avoiding all these new() methods
	protected PVector inOBBSpace(float x_, float y_, float z_) {
		PVector ppos = new PVector(x_, y_, z_);
		PVector rotPpos = new PVector();
		rotPpos = invRotMatrix.mult(ppos, rotPpos);
		PVector innPos = new PVector(rotPpos.x, rotPpos.y, rotPpos.z);		
		innPos.add(position);	
		return innPos;
	}	

	protected float heightOver() {
		return bounds.getExtent().y;
	}

	protected void drawBounds() {
		int fillColor = sandbox.parent.g.fillColor;
		int strokeColor = sandbox.parent.g.strokeColor;
		float strokeWeight = sandbox.parent.g.strokeWeight;		
		
		sandbox.parent.pushMatrix();
		sandbox.parent.stroke(50);    
		sandbox.parent.strokeWeight(1);
		sandbox.parent.noFill();
		sandbox.parent.pushMatrix();
		sandbox.parent.translate(position.x, position.y, position.z);
		sandbox.parent.line(0, 0, 0, heading.x*50, heading.y*50, heading.z*50);
		sandbox.parent.popMatrix();
		sandbox.parent.applyMatrix(rotMatrix);
		PVector extent = bounds.getExtent();
		sandbox.parent.box(2*extent.x, 2*extent.y, 2*extent.z);      
		sandbox.parent.popMatrix();
		
		sandbox.parent.stroke(strokeColor);        
		sandbox.parent.strokeWeight(strokeWeight);		
		sandbox.parent.fill(fillColor);		
	}  	  

}
