package sandbox;

import processing.core.*;
import sandbox.Utils;

import java.util.*; // for ArrayList

public class Sandbox implements PConstants {

    PApplet parent;
    protected boolean isOPENGL;    
    protected HashMap Nodes;
    protected HashMap oldNodes;
    protected HashMap<String, Integer> Collisions;
    protected int countNodes;
    
    // Spatial hashing
    protected float RESOLUTION;
    protected int COLLISION_AREA;
    protected int nX, nY, nZ;
    protected int nX2, nY2, nZ2;
    
    protected int mode;
    /**
     * Constant variables that set sandbox's perspective and camera possibilities.
     * They may be set and changed with {@link sandboxMode()}.
     * Allowed values are FIRSTPERSON and THIRDPERSON.
     *
     * @see http://en.wikipedia.org/wiki/First_person_%28video_games%29
     * @see http://en.wikipedia.org/wiki/Virtual_camera_system
     */
    public final int FIRSTPERSON = 0;
    /**
     * Constant variables that set sandbox's perspective and camera possibilities.
     * They may be set and changed with {@link sandboxMode()}.     
     * Allowed values are FIRSTPERSON and THIRDPERSON.
     *
     * @see http://en.wikipedia.org/wiki/First_person_%28video_games%29
     * @see http://en.wikipedia.org/wiki/Virtual_camera_system
     */    
    public final int THIRDPERSON = 1;  

    
    protected int cameraMode;
    /**
     * Constant variables that set the camera movements.
     * They may be set and changed with {@link cameraMode()}.
     * Allowed values are CENTERED (the controller is shown always at the center
     * of the screen), TRACKING (the controller is seen from behind) and FACING
     * (the controller is seen from the front).     
     *
     * @see http://en.wikipedia.org/wiki/Virtual_camera_system
     */    
    public final int CENTERED = 0;
    /**
     * Constant variables that set the camera movements.
     * They may be set and changed with {@link cameraMode()}.
     * Allowed values are CENTERED (the controller is shown always at the center
     * of the screen), TRACKING (the controller is seen from behind) and FACING
     * (the controller is seen from the front).     
     *
     * @see http://en.wikipedia.org/wiki/Virtual_camera_system
     */        
    public final int TRACKING = 1;
    /**
     * Constant variables that set the camera movements.
     * They may be set and changed with {@link cameraMode()}.
     * Allowed values are CENTERED (the controller is shown always at the center
     * of the screen), TRACKING (the controller is seen from behind) and FACING
     * (the controller is seen from the front).     
     *
     * @see http://en.wikipedia.org/wiki/Virtual_camera_system
     */        
    public final int FACING = 2;
    public final int EYES = 3;        
    
    
    public int controlMode;
    public final int MOUSE_PRESSED = 0;
    public final int MOUSE_SELECT = 1;
    public final int MOUSE_DRAG = 2;
    public final int KEY_PRESSED = 3;
    public final int FREE = 4;
    
    public Camera camera;    
    protected AABB sandboxBounds;
    public int width, height, depth;
    protected PVector orthoInf, orthoSup;
    protected PVector mousePointer;
    protected PVector pointer;    
    protected Node pointedNode;
    protected Node lastNode;
    protected Node lastCollidedNode;
    protected PMatrix3D invCamMatrix;
    public NodeBox defaultGround;
    public int groundColor;    
    protected int gX, gZ;
    
    protected boolean myMousePressed;
    protected boolean myMouseReleased;
    protected boolean myMouseClicked;
    protected boolean myMouseDragged;
    protected boolean myMouseMoved;
    int myMouseX, myMouseY;    
    
    public boolean drawDefaultGround;
    public boolean drawBoundingBoxes;
    public boolean drawCollisionAreas;
    protected int boundsStrokeColor;
    public boolean setWalkable;
    public boolean onFloor;
    public boolean mouseToWalk; // true if mouse is over a walkable node
    public boolean collisions;
    public boolean mouseOver;
    
    public float selfX, selfY, selfZ;
    public float headingX, headingY, headingZ;
    public float mouseX, mouseY, mouseZ;
    public float collisionX, collisionY, collisionZ;
    public float cameraX, cameraY, cameraZ;

    PMatrix3D bindMatrix;
    
    public boolean debug;    
    
    


    /**
     * Main class constructor.
     * TODO: do I hold the AABB wrapping the complete sandbox? Needs to be worked.
     *
     * @param width_
     * @param height_
     * @param depth_
     * @param mode_
     * @param papplet_
     */
    
    public Sandbox(int width_, int height_, int depth_, String mode_, PApplet papplet_) {   
        parent = papplet_;    
        width = width_; height = height_; depth = depth_;
        PVector center = new PVector(width*.5f, -height*.5f, depth*.5f);
        PVector extent = new PVector(width*.5f, height*.5f, depth*.5f);
        sandboxBounds = new AABB(center,extent);        

        // Spatial hashing
        RESOLUTION = 50.f;
        nX = parent.ceil(width/RESOLUTION); nX2 = nX*nX;
        nY = parent.ceil(height/RESOLUTION); nY2 = nY*nY;
        nZ = parent.ceil(depth/RESOLUTION); nZ2 = nZ*nZ;
        
        COLLISION_AREA = 3;
        
        // Initial SANDBOXMODE, CONTROLMODE and CAMERAMODE
        
        controlMode = MOUSE_PRESSED;
        if (mode_ == "FIRSTPERSON") {
            mode = FIRSTPERSON;
            cameraMode = EYES;
        }
        else if (mode_ == "THIRDPERSON") {
            mode = THIRDPERSON;
            cameraMode = CENTERED;
        }
        else mode = FIRSTPERSON;

        // isOpenGL?
        
        processing.core.PGraphics3D pg3d = (PGraphics3D)parent.g;       
         
        if (pg3d.zbuffer != null) isOPENGL = false;
        else isOPENGL = true;
        
        // Node array
        // TODO: oldNodes array necessary?
        
        Nodes = new HashMap();
        oldNodes = new HashMap();
        Collisions = new HashMap();
        countNodes = 1;
        
        // Camera and perspective
        
        camera = new Camera(this);      
        orthoInf = new PVector(-parent.width, -parent.height, -3*parent.height);
        orthoSup = new PVector(parent.width, parent.height, 3*parent.height);
        selfX = camera.controlPos.x;
        selfY = camera.controlPos.y;
        selfZ = camera.controlPos.z;        
        
        // Default ground definition
        
        String hash = spatialHash(center.x, center.y+extent.y, center.z);
        defaultGround = new NodeBox(this, hash);
        defaultGround.setPosition(center.x, center.y+extent.y, center.z);
        defaultGround.setExtent(extent.x, 1, extent.z);
        defaultGround.walkable = true;
        defaultGround.collides = false;  // TODO: supervise: needed in order to move controlPos        
        groundColor = parent.color(100);
        lastNode = defaultGround;        
        gX = 50; gZ = 50;
        
        // WorldModel mouse variables
        
        mousePointer = new PVector();
        mouseX = 0; mouseY = 0; mouseZ = 0;        
        myMousePressed = false;
        myMouseReleased = false;
        myMouseDragged = false;
        myMouseClicked = false;
        myMouseMoved = false;
        myMouseX = 0;
        myMouseY = 0;
        
        // The matrix that will be dynamically attached to nodes with bindBox(), etc. functions
        
        bindMatrix = new PMatrix3D();
        
        // Last collision initial values
        
        collisionX = 0; collisionY = 0; collisionZ = 0;
        
        // Default values for sandbox booleans
        
        drawDefaultGround = true;
        drawBoundingBoxes = false;
        drawCollisionAreas = false;
        boundsStrokeColor = parent.color(0);        
        collisions = false;
        mouseOver = false;
        setWalkable = false;    // TODO this walk() variables...        
        mouseToWalk = false;
        onFloor = true;        
        
        
        parent.registerPre(this);
        parent.registerPost(this);    
        parent.registerMouseEvent(this);
        parent.registerKeyEvent(this);    
        
        debug = false;
    }


    
    
    
    
/////// Public API Methods /////////////////////////////////////////////////////
    
    
    
    
    /**
     *
     * Setups a new sandbox with the specified dimensions and SANDBOXMODE.
     *
     * @see Sandbox()
     *
     */
    
    public static Sandbox createSandbox(int width_, int height_, int depth_, String mode_, PApplet papplet_) {  
        return new Sandbox(width_, height_, depth_, mode_, papplet_);
    }
    
    
    
    
    /**
     * Sets the view mode for the sandbox.
     * Allowed values are FIRSTPERSON or THIRDPERSON
     *
     * @see FIRSTPERSON
     * @see THIRDPERSON
     */        
    public void sandboxMode(int mode_) {
      if (mode_ == 0 || mode_ == 1) {
          mode = mode_;
          
          if (!controlMode(controlMode)) {
              if (mode_ == FIRSTPERSON) controlMode = MOUSE_PRESSED;
              else if (mode_ == THIRDPERSON) controlMode = MOUSE_PRESSED;
          }                    
          if (!cameraMode(cameraMode)) {
              if (mode_ == FIRSTPERSON) cameraMode = EYES;
              else if (mode_ == THIRDPERSON) cameraMode = CENTERED;
          }    
          
          camera.setup();
      }
      else System.err.println("MODE must be FIRSTPERSON or THIRDPERSON");      
    }

    
    
    
    /**
     * Sets the view mode for the sandbox as a String parameter.
     * Allowed values are "FIRSTPERSON" or "THIRDPERSON"
     *
     * @see FIRSTPERSON
     * @see THIRDPERSON
     */
    public void sandboxMode(String mode_) {
        if (mode_ == "FIRSTPERSON") sandboxMode(FIRSTPERSON);
        else if (mode_ == "THIRDPERSON") sandboxMode(THIRDPERSON);
        else System.err.println("MODE must be \"FIRSTPERSON\" or \"THIRDPERSON\"");
    }
    
    
    
    
    /**
     * Retrieves the view mode for the sandbox as a String parameter.
     * Possible values are "FIRSTPERSON" or "THIRDPERSON"
     *
     * @see FIRSTPERSON
     * @see THIRDPERSON
     */    
    public String getSandboxMode() {
        String tmp = "";
        switch(mode) {
        case 0: tmp = "FIRSTPERSON";
        break;
        case 1: tmp = "THIRDPERSON";
        break;
        }
        return tmp;
    }    
    
    
    
    
    /**
     * Sets the cameraMode.
     * Allowed values are CENTERED, TRACKING and FACING
     *
     * @see CENTERED
     * @see TRACKING
     * @see FACING
     */        
    public boolean cameraMode(int mode_) {
        boolean chg = true;
        if (mode_ == parent.constrain(mode_, 0, 3)) {
            if (mode_ == EYES) {
                if (mode == FIRSTPERSON)
                    cameraMode = EYES;
                else {
                    System.err.println("EYES camera mode is only available for FIRSTPERSON mode");
                    chg = false;
                }
            }
            else if (mode_ == CENTERED) {
                if (mode == THIRDPERSON)
                    cameraMode = CENTERED;
                else {
                    System.err.println("CENTERED camera mode is only available for THIRDPERSON mode");
                    chg = false;
                }
            }
            else if (mode_ == FACING) {
                if (controlMode == MOUSE_SELECT || controlMode == FREE)
                    cameraMode = FACING;
                else {
                    System.err.println("FACING camera mode is only available for MOUSE_SELECT or FREE controlModes");
                    chg = false;                  
                }

            }
            else cameraMode = mode_;
        }
        else {
            System.err.println("CAMERAMODE must be CENTERED, TRACKING or FACING");
            chg = false;
        }

        if (chg) camera.setup();
        return chg;
    }

    
    
    
    /**
     * Sets the cameraMode as a String parameter.
     * Allowed values are "CENTERED", "TRACKING" and "FACING"
     *
     * @see CENTERED
     * @see TRACKING
     * @see FACING
     */    
    public boolean cameraMode(String mode_) {
        boolean chg;
        if (mode_ == "CENTERED") chg = cameraMode(CENTERED);
        else if (mode_ == "TRACKING") chg = cameraMode(TRACKING);
        else if (mode_ == "FACING") chg = cameraMode(FACING);
        else if (mode_ == "EYES") chg = cameraMode(EYES);
        else {
            System.err.println("CAMERAMODE must be \"CENTERED\", \"TRACKING\" or \"FACING\"");
            chg = false;
        }
        return chg;
    }

    
    
    
    /**
     * Retrieves the camera mode as a String parameter.
     * Possible values are "CENTER", "TRACKING" or "FACING"
     *
     * @see CENTERED
     * @see TRACKING
     * @see FACING
     */    
    public String getCameraMode() {
        String tmp = "";
        switch(cameraMode) {
        case 0: tmp = "CENTERED";
        break;
        case 1: tmp = "TRACKING";
        break;
        case 2: tmp = "FACING";
        break;        
        case 3: tmp = "EYES";
        break;
        }
        return tmp;
    }
    
    
    
    
    /**
     * Sets the distance between the camera and the point it has been 
attached to
     * for first person sandboxs.
     */        
    public void cameraDistance(float cd) {
        camera.cameraDistance = cd;
    }
    
    
    
    
    /**
     * Sets the controlMode for the movements input.
     * Allowed values are MOUSE_PRESSED, MOUSE_SELECT, MOUSE_DRAG, 
KEY_PRESSED and FREE
     *
     * @see MOUSE_PRESSED
     * @see MOUSE_SELECT
     * @see MOUSE_DRAG
     * @see KEY_PRESSED
     * @see FREE
     */        
    public boolean controlMode(int mode_) {
        boolean chg = true;
        if (mode_ == parent.constrain(mode_, 0, 4)) {
            controlMode = mode_;
            if (!cameraMode(cameraMode)) {
                if (mode_ == FIRSTPERSON) cameraMode = EYES;
                else if (mode_ == THIRDPERSON) cameraMode = CENTERED;
            }
        }
        else {
            System.err.println("CONTROLMODE must be MOUSE_PRESSED, MOUSE_SELECT, MOUSE_DRAG, KEY_PRESSED or FREE.");
            chg = false;
        }
        return chg;
    }
    
    
    
    
    /**
     * Sets the controlMode for the movements input as a String parameter.
     * Allowed values are "MOUSE_PRESSED", "MOUSE_SELECT", "MOUSE_DRAG" 
and "KEY_PRESSED"
     *
     * @see MOUSE_PRESSED
     * @see MOUSE_SELECT
     * @see MOUSE_DRAG
     * @see KEY_PRESSED
     */            
    public boolean controlMode(String mode_) {
        boolean chg;
        if (mode_ == "MOUSE_PRESSED") chg = controlMode(MOUSE_PRESSED);
        else if (mode_ == "MOUSE_SELECT") chg = controlMode(MOUSE_SELECT);
        else if (mode_ == "MOUSE_DRAG") chg = controlMode(MOUSE_DRAG);
        else if (mode_ == "KEY_PRESSED") chg = controlMode(KEY_PRESSED);
        else if (mode_ == "FREE") chg = controlMode(FREE);
        else {
            System.err.println("CONTROLMODE must be \"MOUSE_PRESSED\", \"MOUSE_SELECT\", \"MOUSE_DRAG\", \"KEY_PRESSED\" or \"FREE\".");
            chg = false;
        }
        return chg;
    }
    
    
    
    
    /**
     * Retrieves the controlMode for the movements input as a String 
parameter.
     * Possible values are "MOUSE_PRESSED", "MOUSE_SELECT", "MOUSE_DRAG" 
or "KEY_PRESSED"
     *       
     * @see MOUSE_PRESSED
     * @see MOUSE_SELECT
     * @see MOUSE_DRAG
     * @see KEY_PRESSED
     * @see FREE
     */    
    public String getControlMode() {
        String tmp = "";
        switch(controlMode) {
        case 0: tmp = "MOUSE_PRESSED";
        break;
        case 1: tmp = "MOUSE_SELECT";
        break;
        case 2: tmp = "MOUSE_DRAG";
        break;
        case 3: tmp = "KEY_PRESSED";
        break;                
        }
        return tmp;
    }
    
    
    
    
    /**
     * Whether the sandbox is a first person one or not
     *
     *  @return TRUE or FALSE
     */        
    public boolean isFirstPerson() {
        if (mode == FIRSTPERSON) return true;
        else return false;
    }
    
    
    
    
    
    /**
     * Sets the speed of the camera movements within TRACKING and FACING 
camera modes.
     * Internally, the value acts as an interpolating factor such as f 
in the following
     * equation: x += f * (xNext - x)
     *
     *  @param factor    Ranging from 0 to 1. Defaults to 0.03    
     */        
    public void trackingFactor(float f_) {
        camera.controlInterpolationFactor = f_;
    }
    
    
    
    
    /**
     * Prevents nodes in the sandbox to be gone through by the controller
     */    
    public void collide() {
        collisions = true;
    }
    
    
    
    
    /**
     * Allows nodes in the sandbox to be gone through by the controller
     */        
    public void noCollide() {
        collisions = false;
    }
    
    
    
    
    /**
     * Makes the control character collide with the collide-able nodes 
in the sandbox
     */
    public void controlCollide() {
        camera.controlCollide = true;
    }
    
    
    
    
    /**
     * Makes the control character transparent to the collide-able nodes 
in the sandbox
     */    
    public void noControlCollide() {
        camera.controlCollide = false;
    }
    
    
    
    
    /**
     * Sets the interaction radius of the control character in case it 
collides with other objects
     *
     * @param radius    float: interaction radius
     */
    public void controlRadius(float r) {
        camera.controlRadius = r;
    }
    
    
    
    
    /**
     * Sets true the display boolean of the bounding boxes. If 
arguments, they be
     * treated as stroke color
     */
    public void showBounds() {
        drawBoundingBoxes = true;
        boundsStrokeColor = parent.color(0);
    }

    
    
    
    /**
     * Sets true the display boolean of the bounding boxes. If 
arguments, they be
     * treated as stroke color
     *
     * @param color     stroke color for bounding boxes
     */    
    public void showBounds(int c_) {
        drawBoundingBoxes = true;
        boundsStrokeColor = c_;
    }
    
    
    
    
    /**
     * Sets false the display boolean of the bounding boxes.
     */
    public void noShowBounds() {
        drawBoundingBoxes = false;
        boundsStrokeColor = parent.color(0);
    }
    
    
    
    public void collisionArea(float d) {
    	COLLISION_AREA = parent.floor(d/RESOLUTION);
    }
    
    
    
    /**
     * Attaches an OOB-cube of w width to a point in the space
     *
     * @params w    float: width of the cube
     */
    public void bindBox(float w) {
    	bindBox(w, w, w);
    }
    
    
    
    /**
     * Attaches an OOB of w, h, d dimensions to a point in the space
     *
     * @params w, h, d     float: dimensions of the box
     */
    public void bindBox(float w, float h, float d) {
        bindMatrix = parent.getMatrix(bindMatrix);
        bindMatrix.preApply(invCamMatrix);
        String hash = spatialHash(bindMatrix.m03, bindMatrix.m13, bindMatrix.m23);
        
        NodeBox nodeBox;
        
        String tmpHash = getNode(Nodes, hash);

        if (tmpHash.equals("NO_NODE_FOUND")) {
            nodeBox = new NodeBox(this, hash);
            if (debug) parent.println("Info:    New node created. FCount: " + parent.frameCount);
            if (debug) parent.println("Info:    "+ Nodes.size() + " nodes");
        }
        else {
            nodeBox = (NodeBox)Nodes.get(tmpHash);
            if (!tmpHash.equals(hash) && !Nodes.containsKey(hash)) nodeBox.setHash(hash);
        }
            
        if (!equalMatrices(bindMatrix, nodeBox.rotMatrix)) {
            nodeBox.setMatrix(bindMatrix);        
            nodeBox.setExtent(w*.5f, h*.5f, d*.5f);
        }
        
        nodeBox.inUse = true;
        nodeBox.collides = collisions;        
        lastNode = nodeBox;
        mouseOverNode(nodeBox);
    }
    
    
    

    /**
     * Returns the collision state of lastNode
     */
    public boolean isColliding() {
        boolean isColliding = false;
        boolean output = lastNode.collision;

        
        Iterator iter = Nodes.keySet().iterator();
        int hash0 = Integer.parseInt(lastNode.nodeHash);
        int hash1 = 0, hashDif = 0;
        int i = 0, j = 0, k = 0;
        PVector coll = new PVector();
        
        while (iter.hasNext()) {
            String tmpHash = (String)iter.next();
    
            hash1 = Integer.parseInt(tmpHash);
            hashDif = Math.abs(hash1-hash0);
            i = hashDif%nX;
            j = (hashDif/nX)%nY;
            k = hashDif/(nX*nY);
            
            if (i < COLLISION_AREA && j < COLLISION_AREA && k < COLLISION_AREA) {
            
                Node tmpNode = (Node)Nodes.get(tmpHash);
                
                if (//!equalMatrices(lastNode.rotMatrix, tmpNode.rotMatrix)
                        !lastNode.uuid.equals(tmpNode.uuid)
                        && lastNode.collides && tmpNode.collides
                        && tmpNode.inUse) {
                
                        if (!Collisions.containsKey(lastNode.nodeHash+tmpNode.nodeHash)) {
                            if (tmpNode.intersectsNode(lastNode)) {
                                    isColliding = true;
                                    tmpNode.collision = true;
                                    coll.set(lastNode.position);
                                    coll.set(Utils.interpolateTo(coll, tmpNode.position, .5f));
                            }
                        }
                        else {
                            if (Collisions.get(lastNode.nodeHash+tmpNode.nodeHash) == 1) {
                                isColliding = true;
                                tmpNode.collision = true;
                                coll.set(lastNode.position);
                                coll.set(Utils.interpolateTo(coll, tmpNode.position, .5f));
                            }
                        }
                        Collisions.put(tmpNode.nodeHash+lastNode.nodeHash, (isColliding)?1:0);
                }
//                else if (!isColliding && lastNode.collision && lastNode.uuid.equals(tmpNode.uuid)) parent.println("filtro node");
            }
        }
        if (isColliding) {
            lastNode.collision = true;
            collisionX = coll.x; collisionY = coll.y; collisionZ = coll.z;
        }
        else lastNode.collision = false;
        return output;
    }
    
    
    
    
    /**
     * In MOUSESELECT control mode, attaches a camera to lastNode
     */    
    public void bindCamera() {       // TODO: bindCam ?
        camera.place(lastNode);
    }
    
    
    

    
    /**
     * Sets manually the position of the control character
     *
     * @param x, y, z    float: new position coordinates
     */        
    public void locateSelf(float x_, float y_, float z_) {   // TODO: placeSelf? bindSelf?
        camera.controlPos.set(x_, y_, z_);
        camera.controlPosFut.set(camera.controlPos);
        selfX = (float)camera.controlPos.x;
        selfY = (float)camera.controlPos.y;
        selfZ = (float)camera.controlPos.z;        
    }
    
        
    
    public void firstPersonSpeed(float speed) {
    	camera.vMAX = speed;
    }
    
    
    
    /**
     * TODO: billBoardX(), billBoardY(), billBoardZ()
     *
     * Retrieves bill-boarding angles relative to current position
     */            
    public float billboardX() {
        return 0;
    }

    
    
    
    /**
     *
     * TODO: Fix this ground concept... walk()  still to do... and more
     *
     * Draws default ground as a walk-able rectangle
     *
     */
    public void drawGround() {
        parent.pushMatrix();

        PVector sandboxMin = sandboxBounds.getMin();
        PVector sandboxExtent = sandboxBounds.getExtent();     
        
        parent.pushMatrix();
        parent.translate(sandboxExtent.x, 0, sandboxExtent.z);
        bindMatrix = parent.getMatrix(bindMatrix);
        bindMatrix.preApply(invCamMatrix);
        defaultGround.rotMatrix.set(bindMatrix);        
        parent.popMatrix();
        
        parent.translate(sandboxMin.x, 0, sandboxMin.z);        
        parent.rotateX(parent.HALF_PI);        

        int fillColor = parent.g.fillColor;
        int strokeColor = parent.g.strokeColor;
        float strokeWeight = parent.g.strokeWeight;    
        
        parent.fill(groundColor);
        parent.stroke(groundColor);
        parent.strokeWeight(1);
        
        int rectMode_ = parent.g.rectMode;
        parent.rectMode(CORNER);
        parent.rect(0, 0, 2*sandboxExtent.x, 2*sandboxExtent.z);
        parent.rectMode(rectMode_);
        
        parent.translate(0,-1,0);        
        
        if (drawDefaultGround) {          
            parent.stroke(180);
            for (int x = 0; x < 2*sandboxExtent.x; x += gX)
                parent.line(x, 0, x, 2*sandboxExtent.z);
            for (int z = 0; z < 2*sandboxExtent.z; z += gZ)
                parent.line(0, z, 2*sandboxExtent.x, z);
        }
   
        parent.fill(fillColor);
        parent.stroke(strokeColor);
        parent.strokeWeight(strokeWeight);
        
        parent.popMatrix();    
    }

    
    

    /**
     *
     * TODO: Name of the method, concept still must be thought...
     *
     * Displays the default appearance of the control character while in 
third person
     * view modes.
     *
     */    
    public void drawControl() {
        if (!this.isFirstPerson()) {   
            int fillColor = parent.g.fillColor;
            int strokeColor = parent.g.strokeColor;
            int ellipseMode = parent.g.ellipseMode;
            parent.ellipseMode(CENTER);
            
            parent.pushMatrix();
            parent.translate(this.selfX, -1, this.selfZ);
            parent.rotateX(HALF_PI);
            parent.stroke(0,255,0);
            parent.fill(200,255,200,200);
            parent.ellipse(0, 0, 100, 100);  
            parent.popMatrix();
                
            parent.pushMatrix();
            parent.translate(this.mouseX, -1, this.mouseZ);                
            parent.rotateX(HALF_PI);
            parent.noFill();
            parent.stroke(200,255,200);
            parent.ellipse(0, 0, 100, 100);
            parent.popMatrix();   
            
            parent.ellipseMode(ellipseMode);
            parent.fill(fillColor);
            parent.stroke(strokeColor);            
        }
    }
    

    
    
    /**
     * Initiates the sandbox space, consisting of a view and a camera 
setup. Must be
     * closed with endDraw();
     *
     * TODO: Do I maintain this name?
     */        
    public void beginDraw() {
        cameraSpace();
        parent.pushMatrix();  
   
        float _rotCam = (float)0.;
        if (cameraMode == EYES) {
            if (controlMode == MOUSE_PRESSED) _rotCam = -camera.controlHead.y;
            else if (controlMode == MOUSE_SELECT) _rotCam = camera.controlHead.y;
        }
        else if (cameraMode == TRACKING) {
            if (mode == THIRDPERSON) _rotCam = camera.controlHead.y+(float).5;
            else _rotCam = camera.controlHead.y;
        }
        else if (cameraMode == FACING) {
            if (mode == THIRDPERSON) _rotCam = camera.controlHead.y+(float).5 + PI;
            else _rotCam = camera.controlHead.y;
        }
        
        if (_rotCam != 0.) parent.rotateY(_rotCam);
        parent.translate(-camera.controlPos.x, -camera.controlPos.y, -camera.controlPos.z);
        
        invCamMatrix = parent.getMatrix(invCamMatrix);
        invCamMatrix.invert();
        
        parent.pushMatrix();
        
        drawGround();
    }

    
    

    /**
     * Sets the end of the sandbox space. Must be preceded by a beginDraw() call.
     */    
    public void endDraw() {            
        parent.popMatrix();
        
        if (drawBoundingBoxes) {
            Iterator i = Nodes.values().iterator();
            while (i.hasNext()) {
                Node tmpNode = (Node)i.next();
                tmpNode.drawBounds();
            }          
        }

        if (drawCollisionAreas) {
            Iterator i = Nodes.values().iterator();
            while (i.hasNext()) {
                Node tmpNode = (Node)i.next();
                tmpNode.drawCollisionArea();
            }          
        }
        
        parent.popMatrix();
        mouseUnproject(myMouseX, myMouseY);
    }


    
    
    /**
     * Called before parent.draw().
     */
    public void pre() {
        Collisions.clear();
        oldNodes = (HashMap)Nodes.clone();
        Iterator i = Nodes.values().iterator();
        while (i.hasNext()) {
            Node tmpNode = (Node)i.next();
            if (!tmpNode.inUse) {
                i.remove();
                Nodes.remove(tmpNode.nodeHash);
            }
            tmpNode.inUse = false;
        }
        countNodes = 1;  // It is not 0 because of defaultGround
        defaultGround.inUse = true;
        mouseOverNode(defaultGround);
    }

    
    
    
    /**
     * Called after parent.draw() and after mouse and key events.
     */
    public void post() {  
        if (mode != FIRSTPERSON) camera.isometricSpace();
        else camera.perspectiveSpace();
        
        if (controlMode == MOUSE_PRESSED || controlMode == MOUSE_SELECT)
            mouseOperations();
        
        if (controlMode != FREE) {
            selfX = camera.controlPos.x;
            selfY = camera.controlPos.y;
            selfZ = camera.controlPos.z;
        
            headingX = camera.controlHead.x;
            headingY = camera.controlHead.y;
            headingZ = camera.controlHead.z;
        }
        else {
            camera.controlPos.set(selfX, selfY, selfZ);
            camera.controlHead.set(headingX, headingY, headingZ);
        }        
    }




/////// Protected Methods 
/////////////////////////////////////////////////////

    
    /**
     * Processing perspective() and ortho() functions' setup
     */
    protected void cameraSpace() {       
        if (mode == THIRDPERSON)
            parent.ortho(orthoInf.x, orthoSup.x, orthoInf.y, orthoSup.y, orthoInf.z, orthoSup.z);    
        else {
            parent.perspective();  
            
            int sign = (cameraMode == FACING)?-1:1;
            camera.head(sign*parent.width*(float).5,
                        myMouseY-parent.height*(float).5,
                        sign*(myMouseX-parent.width*(float).5)*3);       
         
        }
        camera.anchor();            
    }



    
    /**
     * Node pool, to avoid continuous object creations and expensive matrix operations
     *
     * @return Node
     */
    protected NodeBox getNodeBox(String hash) {
        NodeBox tmpNode;
        
        if (Nodes.containsKey(hash))
            tmpNode = (NodeBox)Nodes.get(hash);
        else
            tmpNode = new NodeBox(this, hash);
        tmpNode.inUse = true;
        return tmpNode;
    }    
/*
    protected NodeBox getNodeBox() {
        NodeBox tmpNode;
        
        if (Nodes.size() < countNodes+1)
            tmpNode = new NodeBox(this);
        else if (Nodes.get(countNodes) == null)
            tmpNode = new NodeBox(this);    
        else {
            if (((Node)Nodes.get(countNodes)).isBox) tmpNode = 
(NodeBox)Nodes.get(countNodes);
            else tmpNode = new NodeBox(this);
            
            if (tmpNode.inUse) tmpNode = new NodeBox(this);
            tmpNode.inUse = true;
        }
        countNodes++;
        return tmpNode;
    }
*/    
    
    
    
    /**
     * Informs nodes and sandbox if mouse is over a specific node
     *
     * @param node_
     */
    protected void mouseOverNode(Node node_) {
        if (node_.containsVector(mousePointer)) {
            mouseOver = true;
            pointedNode = node_;
        }
        else mouseOver = false;
    }
    

    
    
    /**
     * Informs nodes and sandbox if mouse is over a node of the sandbox
     */
    protected void mouseOver() {
        pointedNode = null;
        
        Iterator i = Nodes.values().iterator();
        while (i.hasNext()) {      
            Node tmpNode = (Node)i.next();
            if (tmpNode.containsVector(mousePointer)) {
                tmpNode.mouseOver = true;
                pointedNode = tmpNode;
            }
            else tmpNode.mouseOver = false;
        }    
        mouseOperations();
    }
    
    

    

    /**
     * Processing mouse events
     */
    protected void mouseOperations() {   
        if (myMousePressed) {
            if (mode != FIRSTPERSON) {      
                mouseToControl();     
            }        
        }
        if (myMouseReleased) {
            if (mode != FIRSTPERSON) {    
                myMouseReleased = false;
                mouseToControl();     
            }        
        }        
        if (myMouseDragged) {
            if (mode != FIRSTPERSON) {      
                mouseToControl();     
            }        
        }
        if (myMouseMoved) {
            myMouseMoved = false;
        }
    }
    


    

    /**
     * Tests if a vector is inside any node in the sandbox
     */
    protected boolean testCollisions(Node testNode_, PVector pos_) {
        boolean out = false;

        Iterator i = Nodes.values().iterator();
        while (i.hasNext()) {   
            Node tmpNode = (Node)i.next();
            if (tmpNode != testNode_) {
                if (tmpNode.walkable && testNode_ == null) {
                    if (tmpNode.containsVector(pos_)) {
                        out = true;
                        lastCollidedNode = tmpNode;
                    }
                }
                else if (tmpNode.collides) {
                    if (testNode_ == null) {
                        if (tmpNode.intersectsSphere(pos_, camera.controlRadius)) {
                            out = true;
                            lastCollidedNode = tmpNode;
                        }
                    }
                    else {
                        if (tmpNode.intersectsNode(testNode_)) {
                            out = true; //TODO OOB
                            lastCollidedNode = tmpNode;
                        }
                    }
                }
            }
        }
        return out;
    }




    /**
     * Utils
     * From 2D screen coordinates to world coordinates
     * Based on function by Seung Joon Choi from his "Organizing Critters" work
     * published in http://www.openprocessing.org/visuals/?visualID=6691
     *
     * @param x
     * @param y
     * @return float[]
     */
    protected float[] unproject(float x, float y) {     
        if ( x < 1 || x >= parent.width-1 || y < 1 || y >= parent.height-1) return null;
        else {
            try {
                float z = 0;
                PMatrix3D modelview, projection;
                
                if (!isOPENGL) {   
                    z = Utils.zBuffer(parent, x, y);
                    modelview = Utils.getModelview(parent); 
                    projection = Utils.getProjection(parent);                 
                }
                else {
                    z = UtilsGL.zBuffer(parent, x, y);
                    modelview = UtilsGL.getModelview(parent);
                    projection = UtilsGL.getProjection(parent);
                }
                float[] in = new float[4];
                float[] out = new float[4];
                in[0] = parent.map(x, 0, parent.width, -1, 1);  
                in[1] = parent.map(y, 0, parent.height, -1, 1);  
                in[2] = parent.map(z, 0, 1, -1, 1);
                in[3] = 1;
                
                modelview.invert();
                projection.invert();
                modelview.apply(projection);
                modelview.mult(in ,out);

                if(out[3] != 0) {
                    out[0] /= out[3];
                    out[1] /= out[3];
                    out[2] /= out[3];
                    return out;
                }
                return null;
            }
            catch(NullPointerException np) {
                System.err.println("No access to z-buffer");
                return null;
            }      
        }
    }

    


    protected void mouseUnproject(int x_, int y_) {
        float tmp[] = unproject((float)x_, (float)y_);
        if (tmp != null) {        
            PVector tmpUnPrj = new PVector(tmp[0], tmp[1], tmp[2]);

            float _rotCam = (float)0.;

            if (cameraMode == EYES)
                _rotCam = -camera.controlHead.y;
            if (cameraMode == TRACKING) {
                if (mode == THIRDPERSON) _rotCam = camera.controlHead.y+(float).5;
                else _rotCam = camera.controlHead.y;
            }
            if (cameraMode == FACING) {
                if (mode == THIRDPERSON) _rotCam = camera.controlHead.y+(float).5 + PI;
                else _rotCam = camera.controlHead.y;
            }
            
/*            
            if (mode == FIRSTPERSON) _rotCam = -camera.controlHead.y;
            if (cameraMode == TRACKING || cameraMode == FACING)
                _rotCam = camera.controlHead.y+(float).5;
*/
            
            if (_rotCam != 0.) 
tmpUnPrj.set(Utils.rotateY(tmpUnPrj,_rotCam));

            AABB expandedAABB = sandboxBounds.copy();
            expandedAABB.setExtent(PVector.mult(sandboxBounds.getExtent(),(float)1.3));
            tmpUnPrj.sub(-camera.controlPos.x, -camera.controlPos.y, -camera.controlPos.z);
            if (expandedAABB.containsVector(PVector.mult(tmpUnPrj, (float).9))) {
                mousePointer.set(tmpUnPrj);
                mouseX = mousePointer.x;
                mouseY = mousePointer.y;
                mouseZ = mousePointer.z;
            }
        }  
    }



    protected void mouseToControl() {
        if (controlMode == MOUSE_PRESSED) {
            mouseToWalk = false;
            if (pointedNode != null && pointedNode.walkable) {
                mouseToWalk = true;
                camera.controlPosFut.set(mousePointer);
                camera.controlPosFut.y = camera.controlPos.y;
                camera.controlPosFut.y = pointedNode.position.y
                                  -pointedNode.heightOver()-1;
            }
        }
    }


    protected boolean equalMatrices(PMatrix3D ma, PMatrix3D mb) {
        float s = Math.abs(ma.m00-mb.m00) + Math.abs(ma.m01-mb.m01) + Math.abs(ma.m02-mb.m02) + Math.abs(ma.m03-mb.m03)
                + Math.abs(ma.m10-mb.m10) + Math.abs(ma.m11-mb.m11) + Math.abs(ma.m12-mb.m12) + Math.abs(ma.m13-mb.m13)
                + Math.abs(ma.m20-mb.m20) + Math.abs(ma.m21-mb.m21) + Math.abs(ma.m22-mb.m22) + Math.abs(ma.m23-mb.m23)
                + Math.abs(ma.m30-mb.m30) + Math.abs(ma.m31-mb.m31) + Math.abs(ma.m32-mb.m32) + Math.abs(ma.m33-mb.m33);
        if (s < .01) return true;
        else return false;
    }
    
    
    // Spatial hashing
    
    protected String spatialHash(float x, float y, float z) {
        int i = gridX(x); int j = gridY(y); int k = gridZ(z);
//        int li = gridX(lx); int lj = gridY(ly); int lk = gridZ(lz);
        return parent.str(i + nX*j + nX*nY*k);
    }
    
    protected int gridX(float x) { return parent.floor(x/RESOLUTION); }
    protected int gridY(float y) { return parent.floor(-y/RESOLUTION); }
    protected int gridZ(float z) { return parent.floor(z/RESOLUTION); }
    
    
    protected String getNode(HashMap Nodes_, String hash) {
        String rtnHash = "NO_NODE_FOUND";
        boolean nodeFound = false;
        int tmpHash0 = Integer.parseInt(hash);
        int tmpHash = 0, n = 0;
        
        if (Nodes_.containsKey(hash)
                && !((Node)Nodes_.get(hash)).inUse) {
            rtnHash = hash;
            nodeFound = true;
        }
            
        while (!nodeFound && n < 6) {
            int dh = 0;
            switch (n) {
            case 0:    case 1:    dh = 1; break;
            case 2:    case 3:    dh = nX; break;
            case 4: case 5: dh = nX*nY; break;
            }
            if (n%2 == 0) tmpHash = tmpHash0+dh;
            else tmpHash = tmpHash0-dh;
            String tmpRtnHash = Integer.toString(tmpHash);
            
            if (Nodes_.containsKey(tmpRtnHash)
                    && !((NodeBox)Nodes_.get(tmpRtnHash)).inUse) {
                nodeFound = true;
                rtnHash = tmpRtnHash;
            }
            n++;
        }
        return rtnHash;
    }
    
    
    
    // Events
    
    
    public void mouseEvent(java.awt.event.MouseEvent event) {  // NOTE: called before post!
        myMouseX = event.getX();
        myMouseY = event.getY();

        switch (event.getID()) {
        case java.awt.event.MouseEvent.MOUSE_PRESSED:
            myMousePressed = true;
            break;

        case java.awt.event.MouseEvent.MOUSE_RELEASED:
            myMousePressed = false;
            myMouseDragged = false;
            myMouseClicked = false;
            myMouseReleased = true;
            break;

        case java.awt.event.MouseEvent.MOUSE_CLICKED:
            myMouseClicked = true;
            break;

        case java.awt.event.MouseEvent.MOUSE_DRAGGED:
            myMouseDragged = true;
            break;

        case java.awt.event.MouseEvent.MOUSE_MOVED:
            myMouseMoved = true;
            break;
        }
    }



    public void keyEvent(java.awt.event.KeyEvent event) {  
        char _keyChar = event.getKeyChar();
        int _keyCode = event.getKeyCode();

        switch (event.getID()) {    
        case java.awt.event.KeyEvent.KEY_RELEASED:
            break;
        case java.awt.event.KeyEvent.KEY_PRESSED:
            break;      
        }
    }
    
    
}
