package sandbox;

import processing.core.PConstants;
import processing.core.PGraphics3D;
import processing.core.PMatrix3D;
import processing.core.PVector;

public class Utils implements PConstants {

	
	Utils() {
	}

	public static float zBuffer(processing.core.PApplet parent, float x, float y) {
		processing.core.PGraphics3D pg3d = (PGraphics3D)parent.g; 
		return pg3d.zbuffer[(int)y * parent.width + (int)x];
	}
	
	
	public static PMatrix3D getModelview(processing.core.PApplet parent) {
		processing.core.PGraphics3D pg3d = (PGraphics3D)parent.g;
		return pg3d.modelview.get(); 
	}
	
	
	public static PMatrix3D getProjection(processing.core.PApplet parent) {
		processing.core.PGraphics3D pg3d = (PGraphics3D)parent.g;
		return pg3d.projection.get(); 
	}
	
	
    public static final boolean isZeroVector(PVector v) {
        return Math.abs(v.x) < EPSILON
                && Math.abs(v.y) < EPSILON
                && Math.abs(v.z) < EPSILON;
    }
    
    public static final PVector interpolateTo(PVector v0, PVector v, float f) {
        return new PVector(v0.x + (v.x-v0.x)*f, v0.y + (v.y-v0.y)*f, v0.z + (v.z-v0.z)* f);
    }
    
	public static final PVector rotateX(PVector v, float ang) {
        float coseno = (float) Math.cos(ang);
        float seno = (float) Math.sin(ang);
        float y = seno * v.y + coseno * v.z;
        float z = coseno * v.z - seno * v.y;
        return new PVector(v.x, y, z);
	}	
	
	public static final PVector rotateY(PVector v, float ang) {
        float coseno = (float) Math.cos(ang);
        float seno = (float) Math.sin(ang);
        float z = seno * v.x + coseno * v.z;
        float x = coseno * v.x - seno * v.z;
        return new PVector(x, v.y, z);
	}
	
	public static final PVector rotateZ(PVector v, float ang) {
        float coseno = (float) Math.cos(ang);
        float seno = (float) Math.sin(ang);
        float y = seno * v.x + coseno * v.y;
        float x = coseno * v.x - seno * v.y;
        return new PVector(x, y, v.z);
	}	
	
	public static final PVector toSpherical(PVector v) {
	    float x = Math.abs(v.x) <= EPSILON ? EPSILON : v.x;
	    float radius = (float) Math.sqrt((x * x) + (v.y * v.y) + (v.z * v.z));
	    float z = (float) Math.asin(v.y / radius);
	    float y = (float) Math.atan(v.z / x) + (x < 0.0 ? PI : 0);  
	    return new PVector(radius, y, z);
	}
	
	public static final PVector toCartesian(PVector v) {
        float r = v.x * (float)Math.cos(v.z);
        float x = r * (float)Math.cos(v.y);
        float y = v.x * (float)Math.sin(v.z);
        float z = r * (float)Math.sin(v.y);
        return new PVector(x, y, z);
	}    
    
    public static final float headingXY(PVector v) {
        return (float) Math.atan2(v.y, v.x);
    } 
    
    public static final float headingXZ(PVector v) {
        return (float) Math.atan2(v.z, v.x);
    }
    
    public static final float headingYZ(PVector v) {
        return (float) Math.atan2(v.y, v.z);
    }     
    
    
	
}
